@@ -17,6 +17,7 @@
 #define HG_PING       	      8
 #define HG_REVLOOKUP 	     16  /* Are we allowed to use the DNS ? */
 #define HG_REVLOOKUP_AS_PING 32 
+#define HG_DISTRIBUTE	     64
 
 struct hg_host {
 	char * hostname;	/* Host name                    */
@@ -40,9 +41,13 @@ struct hg_globals {
 	struct hg_host * host_list;    /* List of tested hosts       */
 	struct hg_host * tested;       /* Tested subnets and domains */
 	int		  flags;       /* options		     */
+	char 		* input;
+	char 		* marker;
+	int	counter;
+	unsigned int	distribute;
 	};
 		 
 struct hg_globals * hg_init(char *, int);
-char * hg_next_host(struct hg_globals *, struct in_addr *);
+int hg_next_host(struct hg_globals *, struct in_addr *, char *, int);
 void   hg_cleanup  (struct hg_globals *);
 #endif
