@@ -13,17 +13,6 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * In addition, as a special exception, Renaud Deraison and Michel Arboi
- * give permission to link the code of this program with any
- * version of the OpenSSL library which is distributed under a
- * license identical to that listed in the included COPYING.OpenSSL
- * file, and distribute linked combinations including the two.
- * You must obey the GNU General Public License in all respects
- * for all of the code used other than OpenSSL.  If you modify
- * this file, you may extend this exception to your version of the
- * file, but you are not obligated to do so.  If you do not wish to
- * do so, delete this exception statement from your version.
- *
  */
 
  
@@ -138,6 +127,24 @@ static struct bpf_pcap * add_pcap(char * iface, pcap_t * pcap)
 }
 
 
+static void rm_pcap(char * iface )
+{
+ struct bpf_pcap * bpc = pcaps, * prev = NULL;
+ while( bpc != NULL )
+ {
+  if(strcmp(bpc->iface, iface) == 0)
+	{
+ 	 struct bpf_pcap * next;
+	 next = bpc->next;
+	 pcap_close(bpc->pcap);
+	 efree(&bpc);
+	 if ( prev ) prev->next = bpc->next;
+	 else pcaps = next;
+	}
+  prev = bpc;
+  bpc = bpc->next;
+ }
+}
 
 static pcap_t * get_pcap(char * iface)
 {
@@ -433,6 +440,7 @@ static int read_clients()
      if(clients[i].iface[0] == '\0')
      { 
        int dl; 
+       int pcap_compile_failed = 0;
        pcap_t * pcap;
        bpf_u_int32 netmask, network;
        if(buf[0] != '\0')buf[strlen(buf) - 1 ] = '\0';
@@ -441,6 +449,7 @@ static int read_clients()
        send(clients[i].soc, ".", 1, 0);
        
        
+again:
        pcap = bpf_add_pcap(clients[i].iface);
        
      
@@ -457,8 +466,21 @@ static int read_clients()
 #endif       
        pcap_lookupnet(clients[i].iface, &network, &netmask, 0);
      /*  pcap_restart(NULL);    */    
-       pcap_compile(pcap, &clients[i].bpf_filter, buf, 0, netmask);
-       clients[i].flag = 1;
+       if ( pcap_compile(pcap, &clients[i].bpf_filter, buf, 0, netmask) < 0 )
+	 {
+	 if ( pcap_compile_failed == 0 )
+		{
+		 rm_pcap(clients[i].iface);
+		 pcap_compile_failed++;
+		 goto again;
+		}
+	  else {
+	   fprintf(stderr, "pcap_compile(%s) failed\n", buf);
+	   send(clients[i].soc, "e", 1, 0);
+           }
+	 }
+        else 
+	 clients[i].flag = 1;
       }
        else send(clients[i].soc, "e", 1, 0);
      }
