@@ -66,6 +66,9 @@ typedef struct {
   int		last_ssl_err;	/* Last SSL error code */
 #endif
  pid_t		pid;		/* Owner - for debugging only */
+
+  char*		buf;		/* NULL if unbuffered */
+  int		bufsz, bufcnt, bufptr;
 } nessus_connection;
 
 /* 
@@ -187,6 +190,8 @@ release_connection_fd(fd)
     
  p = &(connections[fd - NESSUS_FD_OFF]);
 
+ efree(&p->buf);
+
 #ifdef HAVE_SSL
  if (p->ssl != NULL)
   SSL_free(p->ssl);
@@ -627,10 +632,11 @@ set_ids_evasion_mode(args, fp)
      struct arglist	*args;
      nessus_connection	*fp;
 {
- char		*ids_evasion_split = plug_get_key(args, "NIDS/TCP/split");
- char 		*ids_evasion_inject = plug_get_key(args, "NIDS/TCP/inject");
- char 		*ids_evasion_short_ttl = plug_get_key(args,"NIDS/TCP/short_ttl");
- char		*ids_evasion_fake_rst = plug_get_key(args, "NIDS/TCP/fake_rst");
+ struct kb_item ** kb = plug_get_kb(args);
+ char		*ids_evasion_split = kb_item_get_str(kb, "NIDS/TCP/split");
+ char 		*ids_evasion_inject = kb_item_get_str(kb, "NIDS/TCP/inject");
+ char 		*ids_evasion_short_ttl = kb_item_get_str(kb,"NIDS/TCP/short_ttl");
+ char		*ids_evasion_fake_rst = kb_item_get_str(kb, "NIDS/TCP/fake_rst");
  int option = 0;
  
  
@@ -732,13 +738,13 @@ open_stream_connection(args, port, transport, timeout)
   case NESSUS_ENCAPS_SSLv3:
   case NESSUS_ENCAPS_TLSv1:
     renice_myself();
-    cert   = plug_get_key(args, "SSL/cert");
-    key    = plug_get_key(args, "SSL/key");
-    passwd = plug_get_key(args, "SSL/password");
+    cert   = kb_item_get_str(plug_get_kb(args), "SSL/cert");
+    key    = kb_item_get_str(plug_get_kb(args), "SSL/key");
+    passwd = kb_item_get_str(plug_get_kb(args), "SSL/password");
 
-    cafile = plug_get_key(args, "SSL/CA");
+    cafile = kb_item_get_str(plug_get_kb(args), "SSL/CA");
 
-    if ((cafile != NULL) && strlen(cafile))
+    if ((cafile != NULL) && cafile[0] != '\0')
       {
 	cert_names = SSL_load_client_CA_file(cafile);
 	if (cert_names == NULL)
@@ -904,8 +910,8 @@ stream_set_options(fd, reset_opt, set_opt)
 }
 
 
-ExtFunc int 
-read_stream_connection_min(fd, buf0, min_len, max_len)
+static int 
+read_stream_connection_unbuffered(fd, buf0, min_len, max_len)
  int fd;
  void* buf0;
  int min_len, max_len;
@@ -916,6 +922,8 @@ read_stream_connection_min(fd, buf0, min_len, max_len)
   nessus_connection	*fp = NULL;
   fd_set		fdr, fdw;
   struct timeval	tv;
+  time_t		now, then;
+
  
 #if 0
   fprintf(stderr, "read_stream_connection(%d, 0x%x, %d, %d)\n",
@@ -965,17 +973,11 @@ read_stream_connection_min(fd, buf0, min_len, max_len)
 	  if(select(realfd + 1, &fdr, NULL, NULL, timeout > 0 ? &tv : NULL) <= 0)
 	    {
 	      t += INCR_TIMEOUT;
-	      if (min_len <= 0)
-		{
-		  /* Try to be smart */
-		  if (total > 0 && flag) 
-		    return total;
-		  else
-		    flag ++;
-		}
-	      else
-		if (total > min_len)
-		  return total;
+	      /* Try to be smart */
+	      if (total > 0 && flag) 
+		return total;
+	      else if (total >= min_len)
+		flag ++;
 	    }
 	  else
 	    {
@@ -990,6 +992,8 @@ read_stream_connection_min(fd, buf0, min_len, max_len)
 		return total;
 	      /*ret > 0*/
 	      total += ret; 
+	      if (min_len > 0 && total >= min_len)
+		return total;
 	      flag = 0;
 	    }
 	}
@@ -1015,8 +1019,10 @@ read_stream_connection_min(fd, buf0, min_len, max_len)
 
       FD_ZERO(&fdr); FD_ZERO(&fdw);
       FD_SET(realfd, &fdr); FD_SET(realfd, &fdw); 
-      for (t = 0; timeout <= 0 || t < timeout; t += INCR_TIMEOUT)
+      now = then = time(NULL);
+      for (t = 0; timeout <= 0 || t < timeout; t = now - then )
 	{
+          now = time(NULL);
 	  tv.tv_sec = INCR_TIMEOUT; tv.tv_usec = 0;
 	  if (select(realfd+1, &fdr, &fdw, NULL, &tv) > 0)
 	    {
@@ -1072,20 +1078,84 @@ read_stream_connection_min(fd, buf0, min_len, max_len)
 		else
 		  flag ++;
 	      }
-	  else if (total > min_len)
+	  else if (total >= min_len)
 		return total;
 	}
       return total;
 #endif
     default :
-      fprintf(stderr, "Severe bug! Unhandled transport layer %d (fd=%d)\n",
-	      fp->transport, fd);
+      if (fp->transport != -1 || fp->fd != 0)
+	fprintf(stderr, "Severe bug! Unhandled transport layer %d (fd=%d)\n",
+		fp->transport, fd);
+      else
+	fprintf(stderr, "read_stream_connection_unbuffered: fd=%d is closed\n", fd);
       errno = EINVAL;
       return -1;
     }
   /*NOTREACHED*/
 }
 
+ExtFunc int 
+read_stream_connection_min(fd, buf0, min_len, max_len)
+ int fd;
+ void* buf0;
+ int min_len, max_len;
+{
+  nessus_connection	*fp;
+
+  if (NESSUS_STREAM(fd))
+    {
+      fp = &(connections[fd - NESSUS_FD_OFF]);
+      if (fp->buf != NULL)
+	{
+	  int		l1, l2;
+
+	  if (max_len == 1) min_len = 1; /* avoid "magic read" later */
+	  l2 = max_len > fp->bufcnt ? fp->bufcnt : max_len;
+	  if (l2 > 0)
+	    {
+	      memcpy(buf0, fp->buf + fp->bufptr, l2);
+	      fp->bufcnt -= l2;
+	      if (fp->bufcnt == 0)
+		{
+		  fp->bufptr = 0;
+		  fp->buf[0] = '\0'; /* debug */
+		}
+	      else
+		fp->bufptr += l2;
+	      if (l2 >= min_len || l2 >= max_len)
+		return l2;
+	      max_len -= l2;
+	      min_len -= l2;
+	    }
+	  if (min_len > fp->bufsz)
+	    {
+	      l1 = read_stream_connection_unbuffered(fd, (char*)buf0 + l2,
+						     min_len, max_len);
+	      if (l1 > 0)
+		return l1 + l2;
+	      else
+		return l2;
+	    }
+	  /* Fill buffer */
+	  l1 = read_stream_connection_unbuffered(fd, fp->buf, min_len, fp->bufsz);
+	  if (l1 <= 0)
+	    return l2;
+	  
+	  fp->bufcnt = l1;
+	  l1 = max_len > fp->bufcnt ? fp->bufcnt : max_len;
+	  memcpy((char*)buf0 + l2, fp->buf + fp->bufptr, l1);
+	  fp->bufcnt -= l1;
+	  if (fp->bufcnt == 0)
+	    fp->bufptr = 0;
+	  else
+	    fp->bufptr += l1;
+	  return l1 + l2;
+	}
+    }
+  return read_stream_connection_unbuffered(fd, buf0, min_len, max_len);
+}
+
 ExtFunc int 
 read_stream_connection(fd, buf0, len)
  int fd;
@@ -1107,6 +1177,7 @@ write_stream_connection4(fd, buf0, n, i_opt)
  nessus_connection * fp;
   fd_set		fdr, fdw;
   struct timeval	tv;
+  int e;
 
  if(!NESSUS_STREAM(fd))
    {
@@ -1194,8 +1265,14 @@ write_stream_connection4(fd, buf0, n, i_opt)
 		tv.tv_sec = fp->timeout;
      else 
 		tv.tv_sec = TIMEOUT;
+
 	      tv.tv_usec = 0;
-	      if (select(fp->fd+1, &fdr, &fdw, NULL, &tv) <= 0)
+ 	      do {
+ 	      errno = 0;
+	      e = select(fp->fd+1, &fdr, &fdw, NULL, &tv);
+ 	      } while ( e < 0 && errno == EINTR );
+
+	    if ( e <= 0 )
 		{
 #if DEBUG_SSL > 0
 		  nessus_perror("select");
@@ -1207,8 +1284,11 @@ write_stream_connection4(fd, buf0, n, i_opt)
     break;
 #endif
    default:
-     fprintf(stderr, "Severe bug! Unhandled transport layer %d (fd=%d)\n",
-	     fp->transport, fd);
+     if (fp->transport != -1 || fp->fd != 0)
+       fprintf(stderr, "Severe bug! Unhandled transport layer %d (fd=%d)\n",
+	       fp->transport, fd);
+     else
+       fprintf(stderr, "read_stream_connection_unbuffered: fd=%d is closed\n", fd);
      errno =EINVAL;
      return -1;
   }
@@ -1272,7 +1352,7 @@ nsend (fd, data, length, i_opt)
        errno = 0;
        e  = select(fd + 1, NULL, &wr, NULL, &tv);
        if ( e > 0 )
-        n = send(fd, data, length, i_opt);
+        n = os_send(fd, data, length, i_opt);
        else if ( e < 0 && errno == EINTR ) continue;
        else break;
      }
@@ -1288,6 +1368,7 @@ nrecv (fd, data, length, i_opt)
  void * data;
  int length, i_opt;
 {
+  int e;
 #if DEBUG_SSL > 8
    fprintf(stderr, "nrecv: fd=%d len=%d\n", fd, length);
 #endif
@@ -1298,9 +1379,16 @@ nrecv (fd, data, length, i_opt)
   else 
     return read_stream_connection(fd, data, length);
  }
- /* Trying OS's recv() */
- block_socket(fd);		/* ??? */
- return recv(fd, data, length, i_opt);
+ /* Trying OS's recv() 
+  *
+  * Do *NOT* use os_recv() here, as it will be blocking until the exact
+  * amount of requested data arrives
+  */
+ block_socket(fd);
+ do {
+	e = recv(fd, data, length, i_opt);
+ } while ( e < 0 && errno == EINTR );
+ return e;
 }
  
 ExtFunc int
@@ -1544,13 +1632,33 @@ int open_sock_tcp_hn(hostname, port)
 }
 
 
+
 ExtFunc
 int open_sock_tcp(args, port, timeout)
  struct arglist * args; 
  unsigned int port;
  int timeout;
 {
-  return open_sock_option(args, port, SOCK_STREAM,IPPROTO_TCP, timeout);
+  char name[32];
+  int ret;
+  int type;
+  
+
+  /*
+   * If we timed out against this port in the past, there's no need
+   * to scan it again
+   */
+  snprintf(name, sizeof(name), "/tmp/ConnectTimeout/TCP/%d", port);
+  if ( plug_get_key ( args, name, &type ) ) 
+	return -1;
+
+
+  errno = 0;
+  ret  = open_sock_option(args, port, SOCK_STREAM,IPPROTO_TCP, timeout);
+  if ( ret < 0 && errno == ETIMEDOUT )
+    plug_set_key( args, name, ARG_INT, (void*)1); 
+
+  return ret;
 }
 
 
@@ -1658,7 +1766,15 @@ int open_sock_option(args, port, type, protocol, timeout)
   struct sockaddr_in addr;
   struct in_addr * t;
 
+#if 0
+  /* 
+   * MA 2004-08-15: IMHO, as this is often (always?) tested in the NASL scripts
+   * this should not be here. 
+   * If it has to be somewhere else, I'd rather put it in libnasl (and add
+   * a parameter to "force" the connection)
+   */
   if(host_get_port_state(args, port)<=0)return(-1);
+#endif
   bzero((void*)&addr, sizeof(addr));
   addr.sin_family=AF_INET;
   addr.sin_port=htons((unsigned short)port);
@@ -1741,6 +1857,7 @@ int recv_line(soc, buf, bufsiz)
    {
       int e;
  again:
+      errno = 0;
       FD_ZERO(&rd);
       FD_SET(soc, &rd);
       tv.tv_sec = 5;
@@ -1938,3 +2055,198 @@ fd_is_stream(fd)
 {
   return NESSUS_STREAM(fd);	/* Should probably be smarter... */
 }
+
+
+ExtFunc int 
+stream_get_buffer_sz ( int fd )
+{
+  nessus_connection	*p;
+  if (! NESSUS_STREAM(fd))
+    return -1;
+  p = &(connections[fd - NESSUS_FD_OFF]);
+  return p->bufsz;
+}
+
+
+ExtFunc int
+stream_set_buffer(fd, sz)
+     int	fd, sz;
+{
+  nessus_connection	*p;
+  char			*b;
+
+  if (! NESSUS_STREAM(fd))
+    return -1;
+
+  p = &(connections[fd - NESSUS_FD_OFF]);
+  if (sz < p->bufcnt)
+      return -1;		/* Do not want to lose data */
+
+  if (sz == 0)
+    {
+      efree(&p->buf);
+      p->bufsz = 0;
+      return 0;
+    }
+  else if (p->buf == 0)
+    {
+      p->buf = malloc(sz);
+      if (p->buf == NULL)
+	return -1;
+      p->bufsz = sz;
+      p->bufptr = 0;
+      p->bufcnt = 0;
+      return 0;
+    }
+  else
+    {
+      if (p->bufcnt > 0)
+	{
+	  memmove(p->buf, p->buf + p->bufptr, p->bufcnt);
+	  p->bufptr = 0;
+	}
+      b = realloc(p->buf, sz);
+      if (b == NULL)
+	return -1;
+      p->bufsz = sz;
+      return 0;
+    }
+  /*NOTREACHED*/
+}
+
+
+
+/*------------------------------------------------------------------*/
+
+
+int os_send(int soc, void * buf, int len, int opt )
+{
+ char * buf0 = (char*)buf;
+ int e, n;
+ for ( n = 0 ; n < len ; ) 
+ {
+  errno = 0;
+  e = send(soc, buf0 + n , len -  n, opt);
+  if ( e < 0 && errno == EINTR ) continue; 
+  else if ( e <= 0 ) return -1;
+  else n += e;
+ }
+ return n;
+}
+
+int os_recv(int soc, void * buf, int len, int opt )
+{
+ char * buf0 = (char*)buf;
+ int e, n;
+ for ( n = 0 ; n < len ; ) 
+ {
+  errno = 0;
+  e = recv(soc, buf0 + n , len -  n, opt);
+  if ( e < 0 && errno == EINTR ) continue; 
+  else if ( e <= 0 ) return -1;
+  else n += e;
+ }
+ return n;
+}
+
+
+/* 
+ * internal_send() / internal_recv() :
+ *
+ * When processes are passing messages to each other, the format is
+ * <length><msg>, with <length> being a long integer. The functions
+ * internal_send() and internal_recv() encapsulate and decapsulate
+ * the messages themselves. 
+ */
+int internal_send(int soc, char * data, int msg_type )
+{
+ int len;
+ int e;
+ int ack;
+ fd_set rd;
+ struct timeval tv;
+ 
+ if ( data == NULL )
+	data = "";
+
+ e = os_send(soc, &msg_type, sizeof(len), 0 );
+ if ( e < 0 ) return -1;
+
+ if ( (msg_type & INTERNAL_COMM_MSG_TYPE_CTRL) == 0 )
+  {
+ len = strlen(data);
+
+ e = os_send(soc, &len, sizeof(len), 0 );
+ if ( e < 0 ) return -1;
+ e = os_send(soc, data, len, 0 );
+ if ( e < 0 ) return -1;
+ }
+
+ e = os_recv(soc, &ack, sizeof(ack), 0);
+ if ( e < 0 ){
+	fprintf(stderr, "internal_send->os_recv(%d): %s\n",soc, strerror(errno));
+	return -1;
+	}
+
+ return 0;
+}
+
+
+int internal_recv(int soc, char ** data, int * data_sz, int * msg_type )
+{
+ int len = 0;
+ int e;
+ char * buf = *data;
+ int    sz  = *data_sz;
+ fd_set rd;
+ struct timeval tv;
+ int type;
+ int ack;
+ 
+ if ( buf == NULL )
+ {
+  sz = 65535;
+  buf = emalloc ( sz );
+ }
+
+   
+ e = os_recv(soc, &type, sizeof(type), 0 );
+ if ( e < 0 ) goto error;
+
+ if ( (type & INTERNAL_COMM_MSG_TYPE_CTRL) == 0 )
+ {
+ e = os_recv(soc, &len, sizeof(len), 0);
+ if ( e < 0 ) goto error;
+ 
+ if ( len >= sz )
+ {
+  sz = len + 1;
+  buf = erealloc( buf, len );
+ }
+
+ if ( len > 0 )
+ {
+ e = os_recv(soc, buf,len, 0);
+ if ( e < 0 ) goto error;
+ buf[len] = '\0';
+ }
+
+ if ( data != NULL )
+ 	*data = buf;
+ if ( data_sz != NULL )
+ 	*data_sz = sz;
+ }
+ 
+ *msg_type = type;
+ ack = INTERNAL_COMM_MSG_TYPE_CTRL | INTERNAL_COMM_CTRL_ACK;
+ e = os_send(soc, &ack, sizeof(ack), 0);
+ if ( e < 0 ) goto error;
+
+ 
+ return len;
+error:
+ efree(&buf);
+ *data = NULL;
+ *data_sz = 0;
+ return -1;
+}
